@@ -560,7 +560,7 @@ class Sf:
     ### Asynchronous Data Functions
     ### ----------------------------------------------------------------------------------------------------
 
-    async def async_query(self, query) -> pd.DataFrame:
+    async def async_query(self, query, excludedFields: Optional[List] = []) -> pd.DataFrame:
         async_query = async_wrap(self.query)
         return await async_query(query, excludedFields)
     
@@ -688,6 +688,15 @@ class Sf:
         
         return result_dict
     
+    async def field_permission_user_bulk_check(self, objects: list, permissions: list,  users= [], ):
+        data = {}
+        async_field_permission_check = async_wrap(self.field_permission_user_check)
+        result_list = await asyncio.gather(*[async_field_permission_check(object, permission,  users ) for object in objects for permission in permissions])
+        # for result in result_list:
+        #     data = deep_merge_dictionaries(data, result)
+        result_list = [item for sublist in result_list for item in sublist]
+        return result_list
+
     async def entity_access_bulk_query(self, entity_types: list):
         async_entity_access_query = async_wrap(self.entity_access_query)
         results = await asyncio.gather(*[async_entity_access_query(entity_type) for entity_type in entity_types])
@@ -1063,12 +1072,17 @@ class Sf:
 
     def field_permission_check(self, object) -> dict[str, pd.DataFrame]:
         fls_results = transform_sf_result_set_rec(self.sf.query_all(f"""
-                                            SELECT Field,ParentId,PermissionsEdit,PermissionsRead,SobjectType FROM FieldPermissions
+                                            SELECT Field,
+                                            ParentId,
+                                            PermissionsEdit,
+                                            PermissionsRead,
+                                            SobjectType FROM FieldPermissions
                                             WHERE SobjectType in ('{object}')
                                             """)['records'])
 
         permissionSetProfiles = transform_sf_result_set_rec(self.sf.query_all("""
-                                                    SELECT Id,Description,IsOwnedByProfile,Name,ProfileId, Profile.Name, Profile.Description,Type FROM PermissionSet
+                                                    SELECT Id,Description,IsOwnedByProfile,Name,ProfileId, 
+                                                    Profile.Name, Profile.Description,Type FROM PermissionSet
                                                     """)['records'])
 
         profiles = permissionSetProfiles[permissionSetProfiles['PermissionSet.IsOwnedByProfile'] == True][['PermissionSet.ProfileId','PermissionSet.Id','Profile.Name','Profile.Description']].copy()
@@ -1087,6 +1101,64 @@ class Sf:
         
         return {'permissionSetPermissions': permissionSetPermissions, 'profilePermissions': profilePermissions}
 
+    def field_permission_user_check(self, object, permission, users=[]):
+        # Similar to the field_permission_check function, but this one will return a list of users that have the specified permission on the specified object
+        
+        relevant_ps_and_profiles = transform_sf_result_set_rec(self.sf.query(f"""
+                                                SELECT ParentId,
+                                                        Field,
+                                                        PermissionsEdit,
+                                                        PermissionsRead
+                                                FROM FieldPermissions
+                                                WHERE SObjectType IN ('{object}') AND
+                                                ({permission} = true)
+                                                """)['records'])
+
+        relevant_ps_and_profiles['Object API Name'] = relevant_ps_and_profiles['FieldPermissions.Field'].apply(lambda row: row.split('.')[0])
+        relevant_ps_and_profiles['Field API Name'] = relevant_ps_and_profiles['FieldPermissions.Field'].apply(lambda row: row.split('.')[1])
+        relevant_ps_and_profiles.drop(columns=['FieldPermissions.Field'], inplace=True)
+
+        user_query = (" AND Assignee.UserName IN ('" + "','".join(users) + "')") if len(users) > 0 else ""
+
+        relevant_ps_and_profiles_list = "','".join(relevant_ps_and_profiles['FieldPermissions.ParentId'].unique().tolist())
+
+        data = transform_sf_result_set_rec(self.sf.query(f"""SELECT Assignee.Id, Assignee.Name,Assignee.IsActive, 
+                                                                Assignee.UserName, PermissionSet.Id, 
+                                                                PermissionSet.isOwnedByProfile, PermissionSet.Profile.Name, PermissionSet.Label
+                                                                FROM PermissionSetAssignment
+                                                                WHERE PermissionSetId
+                                                                IN ('{relevant_ps_and_profiles_list}')
+                                                                {user_query} AND Assignee.IsActive = TRUE""")['records'])
+
+        data = data.merge(relevant_ps_and_profiles, left_on='PermissionSet.Id', right_on='FieldPermissions.ParentId', how='left').copy()
+
+        if (len(data) > 0):
+                profile_violations = data[data['PermissionSet.IsOwnedByProfile']== True][['User.Username', 'Profile.Name','Field API Name']]
+                permission_set_violations = data[data['PermissionSet.IsOwnedByProfile']== False][['User.Username', 'PermissionSet.Label', 'Field API Name']]
+                result = []
+                for field in profile_violations['Field API Name'].unique():
+                        for user in profile_violations['User.Username'].unique():
+                                result.append({"UserName": user, 
+                                                "Object API Name" : object,
+                                                "Field API Name" : field,
+                                                "Permission": permission,
+                                                "Profile": "; ".join(list(profile_violations[profile_violations['User.Username'] == user]['Profile.Name'].unique())),
+                                                "Permission Set": ""
+                                                
+                                                }
+                                                
+                                                )
+                        for user in permission_set_violations['User.Username'].unique():
+                                result.append({"UserName": user, 
+                                                                                "Object API Name" : object,
+                                                                                "Field API Name" : field,
+                                                                                "Permission": permission,
+                                                                                "Profile": "",
+                                                                                "Permission Set": "; ".join(list(permission_set_violations[permission_set_violations['User.Username'] == user]['PermissionSet.Label'].unique()))})
+                return result
+        else:
+                return []
+            
     # Retrieves the Page Layout by Profile by Record Type data for all objects in the org
     def retrieve_profile_layout_record_type_matrix_all(self, return_pivot_table = False):
         account_page_layouts_bulk = transform_sf_result_set_rec(self.tooling_query_all("""select LayoutId, 
